/* Written by Simon Lacoste-Julien on 2002/07/23 */
/* This is a test for Python Extension.
I'm implementing the Euler solver used in the 522 Simulator.
Interface of use:
>>> import Euler
>>> newx = Euler.step(f,x,t,delta_t) # as before!!!
Where f is an array of (derivative functions) [of the form g(x,t)], x is initial array of values,
t is time, and delta_t is step length (both are double).  That's it!
Magic of Python API... */
/* newx = Euler.step(f,x,t,dt) will simply yield:
newx = x + f(x,t) * dt
*/

#include <Python.h>

/***** 2) Module method definition ********/
static PyObject*
step(PyObject* self, PyObject* args) {

    PyObject* f_array = NULL;       /* will hold f */
    PyObject* f_call = NULL;        /* will hold f[i] (to be called) */
    PyObject* x_array= NULL;        /* will hold array x */
    PyObject* newx = NULL;          /* will hold result (array)*/
    PyObject* xObject = NULL;       /* will hold Python x[i] */
    PyObject* arglist = NULL;       /* will hold arguments for call of f[i] */
    int length;             /* size of f and x */
    int i;                  /* array index */
    double x, x2, t, delta_t;   /* x will hold x[i], x2 will hold result of computations  */

    if (PyArg_ParseTuple(args, "OOdd",
            &f_array, &x_array, &t, &delta_t) == 0) {       /* converts Python -> C */
        /* setting error message: */
        PyErr_SetString(PyExc_RuntimeError, "Wasn't able to parse args for Euler.step fct.");
        return NULL;        /* when returning null, you raise an exception */
    } else {
        Py_INCREF(f_array);
        Py_INCREF(x_array);
        if((PyList_Check(f_array) && PyList_Check(x_array)) == 0) {
            PyErr_SetString(PyExc_TypeError, "Error: f or x weren't list!");
            goto error;
        }

        length = PySequence_Length(f_array);    /* assume x_array has same length... */
        newx = PyList_New(length);              /* initializing the return array */
        Py_INCREF(newx);

        /****** Euler loop ********/
          for(i = 0; i < length; i++) {
            Py_XDECREF(f_call);
            Py_XDECREF(xObject);
            f_call = PyList_GetItem(f_array, i);
            xObject = PyList_GetItem(x_array, i);
            Py_INCREF(f_call);
            Py_INCREF(xObject);

            x = PyFloat_AsDouble(xObject);      /* set x = (double) xObject */

            /* evaluating f[i](x,t) : */
            Py_XDECREF(arglist);
            arglist = Py_BuildValue("(O,d)", x_array, t);   /* building (x,t)  NOTE: x is a vector! */
            Py_INCREF(arglist);
            x2 = PyFloat_AsDouble(PyEval_CallObject(f_call, arglist));  /* x2 = f[i](x,t) */

            /* Euler computation: */
            x2 = x + delta_t * x2;

            /* set result[i] = x2 :*/
            PyList_SetItem(newx, i, Py_BuildValue("d", x2));
        }

        /* cleaning pointers: */
        Py_XDECREF(f_array);
        Py_XDECREF(f_call);
        Py_XDECREF(x_array);
        /* not doing it on newx since we're returning it */
        Py_XDECREF(xObject);
        Py_XDECREF(arglist);

        return newx;
    }
    error:
        /* cleaning memory: */
        Py_XDECREF(f_array);
        Py_XDECREF(f_call);
        Py_XDECREF(x_array);
        Py_XDECREF(newx);
        Py_XDECREF(xObject);
        Py_XDECREF(arglist);

        return NULL;        /* this raises exception */
}

/***** 3) Registration table  **************/
/* this gives the association between the function NAMES (to be used
in Python namespace) to function pointers (addresses) giving the C implementation */

static struct PyMethodDef Euler_methods[] = {
    {"step", step, METH_VARARGS, "Implements Euler solver."},
    /* method name, C func ptr., method of calling fct, method doc */
    {NULL, NULL, 0, NULL}                    /* end of table marker */
};

/***** 4) Initialization function for the module *****/
/* this is called the first time the C module is imported in Python;
it builds the Module's attribute dictionary from the registration
table and create an entry for the C module on the sys.module table */
/* name of fct matters: should be init<moduleName> */
void initEuler() {
    (void) Py_InitModule("Euler", Euler_methods);   /* module name, table ptr */
}
